/*
    RGB Equalizer Filter for VirtualDub (1.01).
		Based on Red/Green/Blue (RGB) Adjustment Filter for VirtualDub.(1.0) by Donald A. Graft
    Copyright (C) 1999-2000 Donald A. Graft
	Copyright (C) 2004 Eugene Khoroshavin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	Eugene Khoroshavin
	fdump@mail.ru
*/

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <crtdbg.h>
#include <math.h>
#include <stdlib.h>

#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#include "resource.h"
#include "filter.h"

///////////////////////////////////////////////////////////////////////////

int RunProc(const FilterActivation *fa, const FilterFunctions *ff);
int StartProc(FilterActivation *fa, const FilterFunctions *ff);
int EndProc(FilterActivation *fa, const FilterFunctions *ff);
long ParamProc(FilterActivation *fa, const FilterFunctions *ff);
int InitProc(FilterActivation *fa, const FilterFunctions *ff);
int ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str);
void ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);
int normpos( int a, int b, int c, int pos);
void normrgb(long *r, long *g, long *b);
UCHAR valtoch(int v, int ch);
int chtoval(UCHAR ch1, UCHAR ch2);
long mmaxc(long *r, long *g, long *b);
void norm1h(int *v);
///////////////////////////////////////////////////////////////////////////
#define M_PI (3.1415926535897932384626433832795) // this is a joke
const char * txt11 = "pre-Gamma R:";
const char * txt12 = "pre-Gamma G:";
const char * txt13 = "pre-Gamma B:";
const char * txt2 = "pre-Saturation:";

const char * txt3 = "R-Levels:";
const char * txt4 = "G-Levels:";
const char * txt5 = "B-Levels:";
const char * txt6 = "Saturation:";

typedef struct MyFilterData {
	IFilterPreview		*ifp;

	int					gamR;
	int					gamG;
	int					gamB;
	int					gamS;

	int					lvR[17];
	int					lvG[17];
	int					lvB[17];
	int					lvS[17];

	int					Rlock;
	int					Glock;
	int					Block;
	int					RGBunion;
	int					keepL;
	int					aveLum;
	int					wavesl;
	int					Slock;
	int					separat;

	int					gamFlg;
	int					rgbFlg;
	int					satFlg;

	unsigned char		rtbl[256];
	unsigned char		gtbl[256];
	unsigned char		btbl[256];

	int					nmfunc;
	long				satp;
	long				incs[256];
	long				incr[256];
	long				incg[256];
	long				incb[256];
} MyFilterData;

////////////////////////////////////////////////////////////////////////////////
void setFlg (MyFilterData *mfd);
void satur(MyFilterData *mfd, long *r, long *g, long *b);
void satpr(MyFilterData *mfd, long *r, long *g, long *b);
void rgbdo(MyFilterData *mfd, long *r, long *g, long *b);
void gamma(MyFilterData *mfd, long *r, long *g, long *b);
///////////////////////////////////////////////////////////////////////////////

bool FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	char mary[149];
	mary[0]		= 34;
	mary[1]		= valtoch(mfd->gamR, 1);
	mary[2]		= valtoch(mfd->gamR, 2);
	mary[3]		= valtoch(mfd->gamG, 1);
	mary[4]		= valtoch(mfd->gamG, 2);
	mary[5]		= valtoch(mfd->gamB, 1);
	mary[6]		= valtoch(mfd->gamB, 2);
	mary[7]		= valtoch(mfd->gamS, 1);
	mary[8]		= valtoch(mfd->gamS, 2);

	mary[9]			= valtoch(mfd->lvR[0], 1);
	mary[10]		= valtoch(mfd->lvR[0], 2);
	mary[11]		= valtoch(mfd->lvR[1], 1);
	mary[12]		= valtoch(mfd->lvR[1], 2);
	mary[13]		= valtoch(mfd->lvR[2], 1);
	mary[14]		= valtoch(mfd->lvR[2], 2);
	mary[15]		= valtoch(mfd->lvR[3], 1);
	mary[16]		= valtoch(mfd->lvR[3], 2);
	mary[17]		= valtoch(mfd->lvR[4], 1);
	mary[18]		= valtoch(mfd->lvR[4], 2);
	mary[19]		= valtoch(mfd->lvR[5], 1);
	mary[20]		= valtoch(mfd->lvR[5], 2);
	mary[21]		= valtoch(mfd->lvR[6], 1);
	mary[22]		= valtoch(mfd->lvR[6], 2);
	mary[23]		= valtoch(mfd->lvR[7], 1);
	mary[24]		= valtoch(mfd->lvR[7], 2);
	mary[25]		= valtoch(mfd->lvR[8], 1);
	mary[26]		= valtoch(mfd->lvR[8], 2);
	mary[27]		= valtoch(mfd->lvR[9], 1);
	mary[28]		= valtoch(mfd->lvR[9], 2);
	mary[29]		= valtoch(mfd->lvR[10], 1);
	mary[30]		= valtoch(mfd->lvR[10], 2);
	mary[31]		= valtoch(mfd->lvR[11], 1);
	mary[32]		= valtoch(mfd->lvR[11], 2);
	mary[33]		= valtoch(mfd->lvR[12], 1);
	mary[34]		= valtoch(mfd->lvR[12], 2);
	mary[35]		= valtoch(mfd->lvR[13], 1);
	mary[36]		= valtoch(mfd->lvR[13], 2);
	mary[37]		= valtoch(mfd->lvR[14], 1);
	mary[38]		= valtoch(mfd->lvR[14], 2);
	mary[39]		= valtoch(mfd->lvR[15], 1);
	mary[40]		= valtoch(mfd->lvR[15], 2);
	mary[41]		= valtoch(mfd->lvR[16], 1);
	mary[42]		= valtoch(mfd->lvR[16], 2);

	mary[43]		= valtoch(mfd->lvG[0], 1);
	mary[44]		= valtoch(mfd->lvG[0], 2);
	mary[45]		= valtoch(mfd->lvG[1], 1);
	mary[46]		= valtoch(mfd->lvG[1], 2);
	mary[47]		= valtoch(mfd->lvG[2], 1);
	mary[48]		= valtoch(mfd->lvG[2], 2);
	mary[49]		= valtoch(mfd->lvG[3], 1);
	mary[50]		= valtoch(mfd->lvG[3], 2);
	mary[51]		= valtoch(mfd->lvG[4], 1);
	mary[52]		= valtoch(mfd->lvG[4], 2);
	mary[53]		= valtoch(mfd->lvG[5], 1);
	mary[54]		= valtoch(mfd->lvG[5], 2);
	mary[55]		= valtoch(mfd->lvG[6], 1);
	mary[56]		= valtoch(mfd->lvG[6], 2);
	mary[57]		= valtoch(mfd->lvG[7], 1);
	mary[58]		= valtoch(mfd->lvG[7], 2);
	mary[59]		= valtoch(mfd->lvG[8], 1);
	mary[60]		= valtoch(mfd->lvG[8], 2);
	mary[61]		= valtoch(mfd->lvG[9], 1);
	mary[62]		= valtoch(mfd->lvG[9], 2);
	mary[63]		= valtoch(mfd->lvG[10], 1);
	mary[64]		= valtoch(mfd->lvG[10], 2);
	mary[65]		= valtoch(mfd->lvG[11], 1);
	mary[66]		= valtoch(mfd->lvG[11], 2);
	mary[67]		= valtoch(mfd->lvG[12], 1);
	mary[68]		= valtoch(mfd->lvG[12], 2);
	mary[69]		= valtoch(mfd->lvG[13], 1);
	mary[70]		= valtoch(mfd->lvG[13], 2);
	mary[71]		= valtoch(mfd->lvG[14], 1);
	mary[72]		= valtoch(mfd->lvG[14], 2);
	mary[73]		= valtoch(mfd->lvG[15], 1);
	mary[74]		= valtoch(mfd->lvG[15], 2);
	mary[75]		= valtoch(mfd->lvG[16], 1);
	mary[76]		= valtoch(mfd->lvG[16], 2);

	mary[77]		= valtoch(mfd->lvB[0], 1);
	mary[78]		= valtoch(mfd->lvB[0], 2);
	mary[79]		= valtoch(mfd->lvB[1], 1);
	mary[80]		= valtoch(mfd->lvB[1], 2);
	mary[81]		= valtoch(mfd->lvB[2], 1);
	mary[82]		= valtoch(mfd->lvB[2], 2);
	mary[83]		= valtoch(mfd->lvB[3], 1);
	mary[84]		= valtoch(mfd->lvB[3], 2);
	mary[85]		= valtoch(mfd->lvB[4], 1);
	mary[86]		= valtoch(mfd->lvB[4], 2);
	mary[87]		= valtoch(mfd->lvB[5], 1);
	mary[88]		= valtoch(mfd->lvB[5], 2);
	mary[89]		= valtoch(mfd->lvB[6], 1);
	mary[90]		= valtoch(mfd->lvB[6], 2);
	mary[91]		= valtoch(mfd->lvB[7], 1);
	mary[92]		= valtoch(mfd->lvB[7], 2);
	mary[93]		= valtoch(mfd->lvB[8], 1);
	mary[94]		= valtoch(mfd->lvB[8], 2);
	mary[95]		= valtoch(mfd->lvB[9], 1);
	mary[96]		= valtoch(mfd->lvB[9], 2);
	mary[97]		= valtoch(mfd->lvB[10], 1);
	mary[98]		= valtoch(mfd->lvB[10], 2);
	mary[99]		= valtoch(mfd->lvB[11], 1);
	mary[100]		= valtoch(mfd->lvB[11], 2);
	mary[101]		= valtoch(mfd->lvB[12], 1);
	mary[102]		= valtoch(mfd->lvB[12], 2);
	mary[103]		= valtoch(mfd->lvB[13], 1);
	mary[104]		= valtoch(mfd->lvB[13], 2);
	mary[105]		= valtoch(mfd->lvB[14], 1);
	mary[106]		= valtoch(mfd->lvB[14], 2);
	mary[107]		= valtoch(mfd->lvB[15], 1);
	mary[108]		= valtoch(mfd->lvB[15], 2);
	mary[109]		= valtoch(mfd->lvB[16], 1);
	mary[110]		= valtoch(mfd->lvB[16], 2);

	mary[111]		= valtoch(mfd->lvS[0], 1);
	mary[112]		= valtoch(mfd->lvS[0], 2);
	mary[113]		= valtoch(mfd->lvS[1], 1);
	mary[114]		= valtoch(mfd->lvS[1], 2);
	mary[115]		= valtoch(mfd->lvS[2], 1);
	mary[116]		= valtoch(mfd->lvS[2], 2);
	mary[117]		= valtoch(mfd->lvS[3], 1);
	mary[118]		= valtoch(mfd->lvS[3], 2);
	mary[119]		= valtoch(mfd->lvS[4], 1);
	mary[120]		= valtoch(mfd->lvS[4], 2);
	mary[121]		= valtoch(mfd->lvS[5], 1);
	mary[122]		= valtoch(mfd->lvS[5], 2);
	mary[123]		= valtoch(mfd->lvS[6], 1);
	mary[124]		= valtoch(mfd->lvS[6], 2);
	mary[125]		= valtoch(mfd->lvS[7], 1);
	mary[126]		= valtoch(mfd->lvS[7], 2);
	mary[127]		= valtoch(mfd->lvS[8], 1);
	mary[128]		= valtoch(mfd->lvS[8], 2);
	mary[129]		= valtoch(mfd->lvS[9], 1);
	mary[130]		= valtoch(mfd->lvS[9], 2);
	mary[131]		= valtoch(mfd->lvS[10], 1);
	mary[132]		= valtoch(mfd->lvS[10], 2);
	mary[133]		= valtoch(mfd->lvS[11], 1);
	mary[134]		= valtoch(mfd->lvS[11], 2);
	mary[135]		= valtoch(mfd->lvS[12], 1);
	mary[136]		= valtoch(mfd->lvS[12], 2);
	mary[137]		= valtoch(mfd->lvS[13], 1);
	mary[138]		= valtoch(mfd->lvS[13], 2);
	mary[139]		= valtoch(mfd->lvS[14], 1);
	mary[140]		= valtoch(mfd->lvS[14], 2);
	mary[141]		= valtoch(mfd->lvS[15], 1);
	mary[142]		= valtoch(mfd->lvS[15], 2);
	mary[143]		= valtoch(mfd->lvS[16], 1);
	mary[144]		= valtoch(mfd->lvS[16], 2);

	mary[145]	= 0x80			| // not use 8th bit
		(mfd->Rlock)			|
		(mfd->Glock		<< 1)	|
		(mfd->Block		<< 2)	|
		(mfd->RGBunion	<< 3)	|
		(mfd->keepL		<< 4)	|
		(mfd->aveLum	<< 5)	|
		(mfd->wavesl	<< 6)	;
	mary[146]	= 0x80			| // not use 8th bit
		(mfd->Slock)			|
		(mfd->separat		<< 1)	;
	
	mary[147] = 34;
	mary[148] = 0;
	_snprintf(buf, buflen, "Config(%s)", mary, 149);


	return true;
}

void ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
	FilterActivation *fa = (FilterActivation *)lpVoid;
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	char **mar = argv[0].asString();
	UCHAR *mary = (UCHAR *)*mar;
	mfd->gamR			= chtoval(mary[0], mary[1]);
	mfd->gamG			= chtoval(mary[2], mary[3]);
	mfd->gamB			= chtoval(mary[4], mary[5]);
	mfd->gamS			= chtoval(mary[6], mary[7]);

	mfd->lvR[0]			= chtoval(mary[8], mary[9]);
	mfd->lvR[1]			= chtoval(mary[10], mary[11]);
	mfd->lvR[2]			= chtoval(mary[12], mary[13]);
	mfd->lvR[3]			= chtoval(mary[14], mary[15]);
	mfd->lvR[4]			= chtoval(mary[16], mary[17]);
	mfd->lvR[5]			= chtoval(mary[18], mary[19]);
	mfd->lvR[6]			= chtoval(mary[20], mary[21]);
	mfd->lvR[7]			= chtoval(mary[22], mary[23]);
	mfd->lvR[8]			= chtoval(mary[24], mary[25]);
	mfd->lvR[9]			= chtoval(mary[26], mary[27]);
	mfd->lvR[10]		= chtoval(mary[28], mary[29]);
	mfd->lvR[11]		= chtoval(mary[30], mary[31]);
	mfd->lvR[12]		= chtoval(mary[32], mary[33]);
	mfd->lvR[13]		= chtoval(mary[34], mary[35]);
	mfd->lvR[14]		= chtoval(mary[36], mary[37]);
	mfd->lvR[15]		= chtoval(mary[38], mary[39]);
	mfd->lvR[16]		= chtoval(mary[40], mary[41]);

	mfd->lvG[0]			= chtoval(mary[42], mary[43]);
	mfd->lvG[1]			= chtoval(mary[44], mary[45]);
	mfd->lvG[2]			= chtoval(mary[46], mary[47]);
	mfd->lvG[3]			= chtoval(mary[48], mary[49]);
	mfd->lvG[4]			= chtoval(mary[50], mary[51]);
	mfd->lvG[5]			= chtoval(mary[52], mary[53]);
	mfd->lvG[6]			= chtoval(mary[54], mary[55]);
	mfd->lvG[7]			= chtoval(mary[56], mary[57]);
	mfd->lvG[8]			= chtoval(mary[58], mary[59]);
	mfd->lvG[9]			= chtoval(mary[60], mary[61]);
	mfd->lvG[10]		= chtoval(mary[62], mary[63]);
	mfd->lvG[11]		= chtoval(mary[64], mary[65]);
	mfd->lvG[12]		= chtoval(mary[66], mary[67]);
	mfd->lvG[13]		= chtoval(mary[68], mary[69]);
	mfd->lvG[14]		= chtoval(mary[70], mary[71]);
	mfd->lvG[15]		= chtoval(mary[72], mary[73]);
	mfd->lvG[16]		= chtoval(mary[74], mary[75]);

	mfd->lvB[0]			= chtoval(mary[76], mary[77]);
	mfd->lvB[1]			= chtoval(mary[78], mary[79]);
	mfd->lvB[2]			= chtoval(mary[80], mary[81]);
	mfd->lvB[3]			= chtoval(mary[82], mary[83]);
	mfd->lvB[4]			= chtoval(mary[84], mary[85]);
	mfd->lvB[5]			= chtoval(mary[86], mary[87]);
	mfd->lvB[6]			= chtoval(mary[88], mary[89]);
	mfd->lvB[7]			= chtoval(mary[90], mary[91]);
	mfd->lvB[8]			= chtoval(mary[92], mary[93]);
	mfd->lvB[9]			= chtoval(mary[94], mary[95]);
	mfd->lvB[10]		= chtoval(mary[96], mary[97]);
	mfd->lvB[11]		= chtoval(mary[98], mary[99]);
	mfd->lvB[12]		= chtoval(mary[100], mary[101]);
	mfd->lvB[13]		= chtoval(mary[102], mary[103]);
	mfd->lvB[14]		= chtoval(mary[104], mary[105]);
	mfd->lvB[15]		= chtoval(mary[106], mary[107]);
	mfd->lvB[16]		= chtoval(mary[108], mary[109]);

	mfd->lvS[0]			= chtoval(mary[110], mary[111]);
	mfd->lvS[1]			= chtoval(mary[112], mary[113]);
	mfd->lvS[2]			= chtoval(mary[114], mary[115]);
	mfd->lvS[3]			= chtoval(mary[116], mary[117]);
	mfd->lvS[4]			= chtoval(mary[118], mary[119]);
	mfd->lvS[5]			= chtoval(mary[120], mary[121]);
	mfd->lvS[6]			= chtoval(mary[122], mary[123]);
	mfd->lvS[7]			= chtoval(mary[124], mary[125]);
	mfd->lvS[8]			= chtoval(mary[126], mary[127]);
	mfd->lvS[9]			= chtoval(mary[128], mary[129]);
	mfd->lvS[10]		= chtoval(mary[130], mary[131]);
	mfd->lvS[11]		= chtoval(mary[132], mary[133]);
	mfd->lvS[12]		= chtoval(mary[134], mary[135]);
	mfd->lvS[13]		= chtoval(mary[136], mary[137]);
	mfd->lvS[14]		= chtoval(mary[138], mary[139]);
	mfd->lvS[15]		= chtoval(mary[140], mary[141]);
	mfd->lvS[16]		= chtoval(mary[142], mary[143]);

	int flgs = mary[144];
	mfd->Rlock			= ((flgs & 0x01) > 0);
	mfd->Glock			= ((flgs & 0x02) > 0);
	mfd->Block			= ((flgs & 0x04) > 0);
	mfd->RGBunion		= ((flgs & 0x08) > 0);
	mfd->keepL			= ((flgs & 0x10) > 0);
	mfd->aveLum			= ((flgs & 0x20) > 0);
	mfd->wavesl			= ((flgs & 0x40) > 0);
	flgs = mary[145];
	mfd->Slock			= ((flgs & 0x01) > 0);
	mfd->separat		= ((flgs & 0x02) > 0);
	
	mfd->nmfunc = IDC_SAT;
	setFlg(mfd);
	mfd->nmfunc = IDC_RC;
	setFlg(mfd);
}

ScriptFunctionDef func_defs[]={
	{ (ScriptFunctionPtr)ScriptConfig, "Config", "0s" },
	{ NULL },
};

CScriptObject script_obj={
	NULL, func_defs
};

struct FilterDefinition filterDef_tutorial = {

	NULL, NULL, NULL,		// next, prev, module
	"RGB Equalizer(1.01)",	// name
	"17-channels equalizer(RGB, Levels, Saturation) + RGB gamma.",
							// desc
	"Eugene Khoroshavin", 		// maker
	NULL,					// private_data
	sizeof(MyFilterData),	// inst_data_size

	InitProc,				// initProc
	NULL,					// deinitProc
	RunProc,				// runProc
	NULL,					// paramProc
	ConfigProc, 			// configProc
	StringProc, 			// stringProc
	StartProc,				// startProc
	EndProc,				// endProc

	&script_obj,			// script_obj
	FssProc,				// fssProc

};

int StartProc(FilterActivation *fa, const FilterFunctions *ff)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	return 0;
}

int EndProc(FilterActivation *fa, const FilterFunctions *ff)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	return 0;
}

///////////////////////////////////////////////////////////////////////////

int RunProc(const FilterActivation *fa, const FilterFunctions *ff)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	const PixDim	width = fa->src.w;
	const PixDim	height = fa->src.h;
	const PixDim	vol = fa->src.w * fa->src.h + fa->src.w;
	Pixel32 *src, *dst;
	int x, y;
	long r, g, b;
	long salign = (fa->src.pitch - width * 4);
	long dalign = (fa->dst.pitch - width * 4);
	src = fa->src.data;
	dst = fa->dst.data;
	for (y = 0; y < height; y++)
	{
		for (x = 0; x < width; x++)
		{
			b = ((UCHAR *)src)[0];
			g = ((UCHAR *)src)[1];
			r = ((UCHAR *)src)[2];

			if (mfd->gamFlg) gamma(mfd, &r, &g, &b);

			if (mfd->gamS) satpr(mfd, &r, &g, &b);

			if (mfd->rgbFlg) rgbdo(mfd, &r, &g, &b);

			if (mfd->satFlg) satur(mfd, &r, &g, &b);

			dst[0] = ((int) r << 16) | ((int) g << 8) | ((int) b);
			src++;
			dst++;
		}
		src	= (Pixel *)((char *)src + salign);
		dst	= (Pixel *)((char *)dst + dalign);
	}
	return 0;
}

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_tutorial;

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
	if (!(fd_tutorial = ff->addFilter(fm, &filterDef_tutorial, sizeof(FilterDefinition))))
		return 1;

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

	return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
	ff->removeFilter(fd_tutorial);
}

int InitProc(FilterActivation *fa, const FilterFunctions *ff) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	mfd->gamR = 0;
	mfd->gamG = 0;
	mfd->gamB = 0;
	mfd->gamS = 0;
	int i;
	for (i = 0; i < 17; i++)
	{
	mfd->lvR[i] = 0;
	mfd->lvG[i] = 0;
	mfd->lvB[i] = 0;
	mfd->lvS[i] = 0;
	}

	mfd->Rlock = false;
	mfd->Glock = false;
	mfd->Block = false;
	mfd->RGBunion = false;
	mfd->keepL = false;
	mfd->aveLum = true;
	mfd->wavesl = false;
	mfd->Slock = false;
	mfd->separat = false;

	mfd->nmfunc = IDC_RC;
	return 0;
}

BOOL CALLBACK ConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
	MyFilterData *mfd = (MyFilterData *)GetWindowLong(hdlg, DWL_USER);
	int mPr=0;
	bool redo = false;
	bool undo = false;
	HWND vhWnd[17];
	int i;
	vhWnd[0] = GetDlgItem(hdlg, IDC_V0);
	vhWnd[1] = GetDlgItem(hdlg, IDC_V1);
	vhWnd[2] = GetDlgItem(hdlg, IDC_V2);
	vhWnd[3] = GetDlgItem(hdlg, IDC_V3);
	vhWnd[4] = GetDlgItem(hdlg, IDC_V4);
	vhWnd[5] = GetDlgItem(hdlg, IDC_V5);
	vhWnd[6] = GetDlgItem(hdlg, IDC_V6);
	vhWnd[7] = GetDlgItem(hdlg, IDC_V7);
	vhWnd[8] = GetDlgItem(hdlg, IDC_V8);
	vhWnd[9] = GetDlgItem(hdlg, IDC_V9);
	vhWnd[10] = GetDlgItem(hdlg, IDC_VA);
	vhWnd[11] = GetDlgItem(hdlg, IDC_VB);
	vhWnd[12] = GetDlgItem(hdlg, IDC_VC);
	vhWnd[13] = GetDlgItem(hdlg, IDC_VD);
	vhWnd[14] = GetDlgItem(hdlg, IDC_VE);
	vhWnd[15] = GetDlgItem(hdlg, IDC_VF);
	vhWnd[16] = GetDlgItem(hdlg, IDC_VV);

	switch(msg) {
		case WM_INITDIALOG:
			SetWindowLong(hdlg, DWL_USER, lParam);
			mfd = (MyFilterData *)lParam;
			HWND hWnd;

			setFlg(mfd);

			hWnd = GetDlgItem(hdlg, IDC_GAMMA);
			SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-100, 100));
			SendMessage(hWnd, TBM_SETTICFREQ, 10 , 0);

			for (i = 0; i < 17; i++) {
				SendMessage(vhWnd[i], TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-100, 100));
				SendMessage(vhWnd[i], TBM_SETTICFREQ, 10 , 0);
			}

			SendMessage(GetDlgItem(hdlg, IDC_RLOCK), BM_SETCHECK, mfd->Rlock, 0);
			SendMessage(GetDlgItem(hdlg, IDC_GLOCK), BM_SETCHECK, mfd->Glock, 0);
			SendMessage(GetDlgItem(hdlg, IDC_BLOCK), BM_SETCHECK, mfd->Block, 0);
			SendMessage(GetDlgItem(hdlg, IDC_SLOCK), BM_SETCHECK, mfd->Slock, 0);
			SendMessage(GetDlgItem(hdlg, IDC_LAVE), BM_SETCHECK, mfd->aveLum, 0);
			SendMessage(GetDlgItem(hdlg, IDC_LMAX), BM_SETCHECK, (!mfd->aveLum && !mfd->separat), 0);
			SendMessage(GetDlgItem(hdlg, IDC_SEPARAT), BM_SETCHECK, mfd->separat, 0);
			SendMessage(GetDlgItem(hdlg, IDC_KEEPL), BM_SETCHECK, mfd->keepL, 0);
			SendMessage(GetDlgItem(hdlg, IDC_STICK), BM_SETCHECK, mfd->RGBunion, 0);
			SendMessage(GetDlgItem(hdlg, IDC_WAVE), BM_SETCHECK, mfd->wavesl, 0);

			SendMessage(GetDlgItem(hdlg, mfd->nmfunc), BM_CLICK, TRUE, 0);

			mfd->ifp->InitButton(GetDlgItem(hdlg, IDPREVIEW));
			mfd->ifp->Toggle(hdlg);
			return TRUE;

		case WM_COMMAND:
			switch(LOWORD(wParam)) {
			case IDPREVIEW:
				mfd->ifp->Toggle(hdlg);
				break;
			case IDOK:
				EndDialog(hdlg, 0);
				return TRUE;
			case IDHELP:
				{
				char prog[256];
				char path[256];
				LPTSTR ptr;
				GetModuleFileName(NULL, prog, 255);
				GetFullPathName(prog, 255, path, &ptr);
				*ptr = 0;
				strcat(path, "plugins\\rgbEQ.txt");
				OutputDebugString(path);
				OutputDebugString("\n");
				strcpy(prog, "Notepad ");
				strcat(prog, path);
				WinExec(prog, SW_SHOW);
				return TRUE;
				}
			case IDCANCEL:
				EndDialog(hdlg, 1);
				return TRUE;
			case IDC_ZGAMMA:
				switch (mfd->nmfunc) {
				case IDC_RC:
					if (mfd->Rlock) undo = true;
					else if (mfd->RGBunion) {
							mfd->gamR = 0;
							mfd->gamB = 0;
							mfd->gamG = 0;
					}
					else mfd->gamR = 0;
					break;
				case IDC_GC:
					if (mfd->Glock) undo = true;
					else if (mfd->RGBunion) {
							mfd->gamR = 0;
							mfd->gamB = 0;
							mfd->gamG = 0;
					}
					else mfd->gamG = 0;
					break;
				case IDC_BC:
					if (mfd->Block) undo = true;
					else if (mfd->RGBunion) {
							mfd->gamR = 0;
							mfd->gamB = 0;
							mfd->gamG = 0;
					}
					else mfd->gamB = 0;
					break;
				case IDC_SAT:
					if (mfd->Slock) undo = true;
					else mfd->gamS = 0;
					break;
				}
				if (!undo) {
					SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_SETPOS, (WPARAM)TRUE, 0);
					setFlg(mfd);
					mfd->ifp->RedoFrame();
				}
				undo = false;
				return TRUE;
			case IDC_ZEQ:
				if (mfd->RGBunion && (mfd->nmfunc != IDC_SAT)) {
					for (i = 0; i < 17; i++)
					{
					mfd->lvR[i] = 0;
					mfd->lvG[i] = 0;
					mfd->lvB[i] = 0;
					mfd->lvS[i] = 0;
					}
				}
				else {
					switch (mfd->nmfunc) {
					case IDC_RC:
						if (mfd->Rlock) undo = true;
						else {
							for (i = 0; i < 17; i++)
							{
							mfd->lvR[i] = 0;
							}
						}
						break;
					case IDC_GC:
						if (mfd->Glock) undo = true;
						else {
							for (i = 0; i < 17; i++)
							{
							mfd->lvG[i] = 0;
							}
						}
						break;
					case IDC_BC:
						if (mfd->Block) undo = true;
						else {
							for (i = 0; i < 17; i++)
							{
							mfd->lvB[i] = 0;
							}
						}
						break;
					case IDC_SAT:
						if (mfd->Slock) undo = true;
						else {
							for (i = 0; i < 17; i++)
							{
							mfd->lvS[i] = 0;
							}
						}
						break;
					}
				}
				if ( !undo) {
				for (i = 0; i < 17; i++) {
					SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, 0);
				}
					setFlg(mfd);
					mfd->ifp->RedoFrame();
				}
				undo = false;
				return TRUE;
			case IDC_ZALL:
				mfd->Rlock = false;
				mfd->Glock = false;
				mfd->Block = false;
				mfd->RGBunion = false;
				mfd->keepL = false;
				mfd->aveLum = true;
				mfd->wavesl = false;
				mfd->Slock = false;
				mfd->separat = false;
				SendMessage(GetDlgItem(hdlg, IDC_RLOCK), BM_SETCHECK, mfd->Rlock, 0);
				SendMessage(GetDlgItem(hdlg, IDC_GLOCK), BM_SETCHECK, mfd->Glock, 0);
				SendMessage(GetDlgItem(hdlg, IDC_BLOCK), BM_SETCHECK, mfd->Block, 0);
				SendMessage(GetDlgItem(hdlg, IDC_SLOCK), BM_SETCHECK, mfd->Slock, 0);
				SendMessage(GetDlgItem(hdlg, IDC_LAVE), BM_SETCHECK, mfd->aveLum, 0);
				SendMessage(GetDlgItem(hdlg, IDC_LMAX), BM_SETCHECK, (!mfd->aveLum && !mfd->separat), 0);
				SendMessage(GetDlgItem(hdlg, IDC_SEPARAT), BM_SETCHECK, mfd->separat, 0);
				SendMessage(GetDlgItem(hdlg, IDC_KEEPL), BM_SETCHECK, mfd->keepL, 0);
				SendMessage(GetDlgItem(hdlg, IDC_STICK), BM_SETCHECK, mfd->RGBunion, 0);
				SendMessage(GetDlgItem(hdlg, IDC_WAVE), BM_SETCHECK, mfd->wavesl, 0);

				mfd->gamR = 0;
				mfd->gamG = 0;
				mfd->gamB = 0;
				mfd->gamS = 0;

				for (i = 0; i < 17; i++)
				{
				mfd->lvR[i] = 0;
				mfd->lvG[i] = 0;
				mfd->lvB[i] = 0;
				mfd->lvS[i] = 0;
				}
				SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_SETPOS, (WPARAM)TRUE, 0);
				for (i = 0; i < 17; i++) {
					SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, 0);
				}

				setFlg(mfd);
				mfd->ifp->RedoFrame();
				return TRUE;
			case IDC_RC:
				SendMessage(GetDlgItem(hdlg, IDC_HORSLADER), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)txt11);
				SendMessage(GetDlgItem(hdlg, IDC_VERSLADERS), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)txt3);

				SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_SETPOS, (WPARAM)TRUE, mfd->gamR);
				for (i = 0; i < 17; i++) {
					SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -mfd->lvR[i]);
				}
				mfd->nmfunc = IDC_RC;
				return TRUE;
			case IDC_GC:
				SendMessage(GetDlgItem(hdlg, IDC_HORSLADER), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)txt12);
				SendMessage(GetDlgItem(hdlg, IDC_VERSLADERS), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)txt4);

				SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_SETPOS, (WPARAM)TRUE, mfd->gamG);
				for (i = 0; i < 17; i++) {
					SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -mfd->lvG[i]);
				}
				mfd->nmfunc = IDC_GC;
				return TRUE;
			case IDC_BC:
				SendMessage(GetDlgItem(hdlg, IDC_HORSLADER), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)txt13);
				SendMessage(GetDlgItem(hdlg, IDC_VERSLADERS), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)txt5);

				SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_SETPOS, (WPARAM)TRUE, mfd->gamB);
				for (i = 0; i < 17; i++) {
					SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -mfd->lvB[i]);
				}
				mfd->nmfunc = IDC_BC;
				return TRUE;
			case IDC_SAT:
				SendMessage(GetDlgItem(hdlg, IDC_HORSLADER), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)txt2);
				SendMessage(GetDlgItem(hdlg, IDC_VERSLADERS), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)txt6);

				SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_SETPOS, (WPARAM)TRUE, mfd->gamS);
				for (i = 0; i < 17; i++) {
					SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -mfd->lvS[i]);
				}
				mfd->nmfunc = IDC_SAT;
				return TRUE;
			case IDC_LAVE:
			case IDC_LMAX:
			case IDC_SEPARAT:
				mfd->aveLum = IsDlgButtonChecked(hdlg, IDC_LAVE);
				mfd->separat = IsDlgButtonChecked(hdlg, IDC_SEPARAT);
				setFlg(mfd);
				mfd->ifp->RedoFrame();
				return TRUE;
			case IDC_KEEPL:
				mfd->keepL = IsDlgButtonChecked(hdlg, IDC_KEEPL);
				setFlg(mfd);
				mfd->ifp->RedoFrame();
				return TRUE;
			case IDC_RLOCK:
			case IDC_GLOCK:
			case IDC_BLOCK:
				if (IsDlgButtonChecked(hdlg, LOWORD(wParam))) {
				SendMessage(GetDlgItem(hdlg, IDC_STICK), BM_SETCHECK, FALSE, 0);
				mfd->RGBunion = FALSE;
				}
				mfd->Rlock = IsDlgButtonChecked(hdlg, IDC_RLOCK);
				mfd->Glock = IsDlgButtonChecked(hdlg, IDC_GLOCK);
				mfd->Block = IsDlgButtonChecked(hdlg, IDC_BLOCK);
				return TRUE;
			case IDC_SLOCK:
				mfd->Slock = IsDlgButtonChecked(hdlg, IDC_SLOCK);
				return TRUE;
			case IDC_WAVE:
				mfd->wavesl = IsDlgButtonChecked(hdlg, IDC_WAVE);
				return TRUE;
			case IDC_STICK:
				if (mfd->Rlock || mfd->Glock || mfd->Block) {
					SendMessage(GetDlgItem(hdlg, IDC_STICK), BM_SETCHECK, FALSE, 0);
					mfd->RGBunion = FALSE;
				}
				else mfd->RGBunion = IsDlgButtonChecked(hdlg, IDC_STICK);
				return TRUE;
			}
			break;
		case WM_HSCROLL:
			{
				int vv, rold, gold, bold, dpos;
				vv = SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_GETPOS, 0, 0);
				rold = mfd->gamR;
				gold = mfd->gamG;
				bold = mfd->gamB;
				switch (mfd->nmfunc) {
				case IDC_RC:
					if (mfd->Rlock) SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_SETPOS, (WPARAM)TRUE, rold);
					else {
						if (mfd->RGBunion) {
							dpos = vv - rold;
							dpos = normpos(rold, gold, bold, dpos);
							mfd->gamR += dpos;
							mfd->gamG += dpos;
							mfd->gamB += dpos;
						}
						else mfd->gamR = vv;
						SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_SETPOS, (WPARAM)TRUE, mfd->gamR);
					}
					break;
				case IDC_GC:
					if (mfd->Glock) SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_SETPOS, (WPARAM)TRUE, gold);
					else {
						if (mfd->RGBunion) {
							dpos = vv - gold;
							dpos = normpos(rold, gold, bold, dpos);
							mfd->gamR += dpos;
							mfd->gamG += dpos;
							mfd->gamB += dpos;
						}
						else mfd->gamG = vv;
						SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_SETPOS, (WPARAM)TRUE, mfd->gamG);
					}
					break;
				case IDC_BC:
					if (mfd->Block) SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_SETPOS, (WPARAM)TRUE, bold);
					else {
						if (mfd->RGBunion) {
							dpos = vv - bold;
							dpos = normpos(rold, gold, bold, dpos);
							mfd->gamR += dpos;
							mfd->gamG += dpos;
							mfd->gamB += dpos;
						}
						else mfd->gamB = vv;
						SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_SETPOS, (WPARAM)TRUE, mfd->gamB);
					}
					break;
				case IDC_SAT:
					if (mfd->Slock) SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_SETPOS, (WPARAM)TRUE, mfd->gamS);
					else {
						mfd->gamS = vv;
						SendMessage(GetDlgItem(hdlg, IDC_GAMMA), TBM_SETPOS, (WPARAM)TRUE, mfd->gamS);
					}
					break;
				}
			}
			setFlg(mfd);
			mfd->ifp->RedoFrame();
			break;
		case WM_VSCROLL:
			{
				int vv, dpos, slider;
				for (i=0; i<17; i++) {
					if(vhWnd[i] == (HWND)lParam) slider = i;
				}
				switch (mfd->nmfunc) {
				case IDC_RC:
					if (mfd->Rlock) {
						SendMessage((HWND)lParam, TBM_SETPOS, (WPARAM)TRUE, -mfd->lvR[slider]);
						return true;
					}
					break;
				case IDC_GC:
					if (mfd->Glock) {
						SendMessage((HWND)lParam, TBM_SETPOS, (WPARAM)TRUE, -mfd->lvG[slider]);
						return true;
					}
					break;
				case IDC_BC:
					if (mfd->Block) {
						SendMessage((HWND)lParam, TBM_SETPOS, (WPARAM)TRUE, -mfd->lvB[slider]);
						return true;
					}
					break;
				case IDC_SAT:
					if (mfd->Slock) {
						SendMessage((HWND)lParam, TBM_SETPOS, (WPARAM)TRUE, -mfd->lvS[slider]);
						return true;
					}
					break;
				}

				if ((mfd->nmfunc !=IDC_SAT) && mfd->RGBunion) {
					vv = -SendMessage((HWND)lParam, TBM_GETPOS, 0, 0);
					switch (mfd->nmfunc) {
					case IDC_RC:
						if (mfd->wavesl) {
							switch ((int)LOWORD(wParam)) {
							case SB_THUMBTRACK: {
								int pos;
								dpos = vv - mfd->lvR[slider];
								for (i = slider + 1; i < slider + 4 && i < 17; i++)
								{
									pos = mfd->lvR[i] + (int)(dpos * cos(M_PI * (float)((i - slider) / 8.0)));
									norm1h(&pos);
									SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -pos);
								}
								for (i = slider - 1; i > slider - 4 && i >= 0; i--)
								{
									pos = mfd->lvR[i] + (int)(dpos * cos(M_PI * (float)((slider - i) / 8.0)));
									norm1h(&pos);
									SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -pos);
								}
												}
								break;
							case SB_THUMBPOSITION: {
								int pos;
								for (i = slider; (i < (slider + 4)) && (i < 17); i++)
								{
									pos = -SendMessage(vhWnd[i], TBM_GETPOS, 0, 0);
									dpos = pos - mfd->lvR[i];
									dpos = normpos(mfd->lvR[i], mfd->lvG[i], mfd->lvB[i], dpos);
									mfd->lvR[i] += dpos;
									mfd->lvG[i] += dpos;
									mfd->lvB[i] += dpos;
									SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -mfd->lvR[i]);
								}
								for (i = slider - 1; (i > (slider - 4)) && (i >= 0); i--)
								{
									pos = -SendMessage(vhWnd[i], TBM_GETPOS, 0, 0);
									dpos = pos - mfd->lvR[i];
									dpos = normpos(mfd->lvR[i], mfd->lvG[i], mfd->lvB[i], dpos);
									mfd->lvR[i] += dpos;
									mfd->lvG[i] += dpos;
									mfd->lvB[i] += dpos;
									SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -mfd->lvR[i]);
								}
												}
								break;
							case SB_BOTTOM:
							case SB_LINEDOWN:
							case SB_LINEUP:
							case SB_PAGEDOWN:
							case SB_PAGEUP:
							case SB_TOP:
								{
								vv = -SendMessage((HWND)lParam, TBM_GETPOS, 0, 0);
								dpos = vv - mfd->lvR[slider];
								dpos = normpos(mfd->lvR[slider], mfd->lvG[slider], mfd->lvB[slider], dpos);
								mfd->lvR[slider] += dpos;
								mfd->lvG[slider] += dpos;
								mfd->lvB[slider] += dpos;
								SendMessage((HWND)lParam, TBM_SETPOS, (WPARAM)TRUE, -mfd->lvR[slider]);
									 }
								break;
							}
						}
						else {
							dpos = vv - mfd->lvR[slider];
							dpos = normpos(mfd->lvR[slider], mfd->lvG[slider], mfd->lvB[slider], dpos);
							mfd->lvR[slider] += dpos;
							mfd->lvG[slider] += dpos;
							mfd->lvB[slider] += dpos;
							SendMessage((HWND)lParam, TBM_SETPOS, (WPARAM)TRUE, -mfd->lvR[slider]);
						}
						break;
					case IDC_GC:
						if (mfd->wavesl) {
							switch ((int)LOWORD(wParam)) {
							case SB_THUMBTRACK: {
								int pos;
								dpos = vv - mfd->lvG[slider];
								for (i = slider + 1; i < slider + 4 && i < 17; i++)
								{
									pos = mfd->lvG[i] + (int)(dpos * cos(M_PI * (float)((i - slider) / 8.0)));
									norm1h(&pos);
									SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -pos);
								}
								for (i = slider - 1; i > slider - 4 && i >= 0; i--)
								{
									pos = mfd->lvG[i] + (int)(dpos * cos(M_PI * (float)((slider - i) / 8.0)));
									norm1h(&pos);
									SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -pos);
								}
												}
								break;
							case SB_THUMBPOSITION: {
								int pos;
								for (i = slider; (i < (slider + 4)) && (i < 17); i++)
								{
									pos = -SendMessage(vhWnd[i], TBM_GETPOS, 0, 0);
									dpos = pos - mfd->lvG[i];
									dpos = normpos(mfd->lvR[i], mfd->lvG[i], mfd->lvB[i], dpos);
									mfd->lvR[i] += dpos;
									mfd->lvG[i] += dpos;
									mfd->lvB[i] += dpos;
									SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -mfd->lvG[i]);
								}
								for (i = slider - 1; (i > (slider - 4)) && (i >= 0); i--)
								{
									pos = -SendMessage(vhWnd[i], TBM_GETPOS, 0, 0);
									dpos = pos - mfd->lvG[i];
									dpos = normpos(mfd->lvR[i], mfd->lvG[i], mfd->lvB[i], dpos);
									mfd->lvR[i] += dpos;
									mfd->lvG[i] += dpos;
									mfd->lvB[i] += dpos;
									SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -mfd->lvG[i]);
								}
												}
								break;
							case SB_BOTTOM:
							case SB_LINEDOWN:
							case SB_LINEUP:
							case SB_PAGEDOWN:
							case SB_PAGEUP:
							case SB_TOP:
								{
								vv = -SendMessage((HWND)lParam, TBM_GETPOS, 0, 0);
								dpos = vv - mfd->lvG[slider];
								dpos = normpos(mfd->lvR[slider], mfd->lvG[slider], mfd->lvB[slider], dpos);
								mfd->lvR[slider] += dpos;
								mfd->lvG[slider] += dpos;
								mfd->lvB[slider] += dpos;
								SendMessage((HWND)lParam, TBM_SETPOS, (WPARAM)TRUE, -mfd->lvG[slider]);
									 }
								break;
							}
						}
						else {
							dpos = vv - mfd->lvG[slider];
							dpos = normpos(mfd->lvR[slider], mfd->lvG[slider], mfd->lvB[slider], dpos);
							mfd->lvR[slider] += dpos;
							mfd->lvG[slider] += dpos;
							mfd->lvB[slider] += dpos;
							SendMessage((HWND)lParam, TBM_SETPOS, (WPARAM)TRUE, -mfd->lvG[slider]);
						}
						break;
					case IDC_BC:
						if (mfd->wavesl) {
							switch ((int)LOWORD(wParam)) {
							case SB_THUMBTRACK: {
								int pos;
								dpos = vv - mfd->lvB[slider];
								for (i = slider + 1; i < slider + 4 && i < 17; i++)
								{
									pos = mfd->lvB[i] + (int)(dpos * cos(M_PI * (float)((i - slider) / 8.0)));
									norm1h(&pos);
									SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -pos);
								}
								for (i = slider - 1; i > slider - 4 && i >= 0; i--)
								{
									pos = mfd->lvB[i] + (int)(dpos * cos(M_PI * (float)((slider - i) / 8.0)));
									norm1h(&pos);
									SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -pos);
								}
												}
								break;
							case SB_THUMBPOSITION: {
								int pos;
								for (i = slider; (i < (slider + 4)) && (i < 17); i++)
								{
									pos = -SendMessage(vhWnd[i], TBM_GETPOS, 0, 0);
									dpos = pos - mfd->lvB[i];
									dpos = normpos(mfd->lvR[i], mfd->lvG[i], mfd->lvB[i], dpos);
									mfd->lvR[i] += dpos;
									mfd->lvG[i] += dpos;
									mfd->lvB[i] += dpos;
									SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -mfd->lvB[i]);
								}
								for (i = slider - 1; (i > (slider - 4)) && (i >= 0); i--)
								{
									pos = -SendMessage(vhWnd[i], TBM_GETPOS, 0, 0);
									dpos = pos - mfd->lvB[i];
									dpos = normpos(mfd->lvR[i], mfd->lvG[i], mfd->lvB[i], dpos);
									mfd->lvR[i] += dpos;
									mfd->lvG[i] += dpos;
									mfd->lvB[i] += dpos;
									SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -mfd->lvB[i]);
								}
												}
								break;
							case SB_BOTTOM:
							case SB_LINEDOWN:
							case SB_LINEUP:
							case SB_PAGEDOWN:
							case SB_PAGEUP:
							case SB_TOP:
								{
								vv = -SendMessage((HWND)lParam, TBM_GETPOS, 0, 0);
								dpos = vv - mfd->lvB[slider];
								dpos = normpos(mfd->lvR[slider], mfd->lvG[slider], mfd->lvB[slider], dpos);
								mfd->lvR[slider] += dpos;
								mfd->lvG[slider] += dpos;
								mfd->lvB[slider] += dpos;
								SendMessage((HWND)lParam, TBM_SETPOS, (WPARAM)TRUE, -mfd->lvB[slider]);
									 }
								break;
							}
						}
						else {
							dpos = vv - mfd->lvB[slider];
							dpos = normpos(mfd->lvR[slider], mfd->lvG[slider], mfd->lvB[slider], dpos);
							mfd->lvR[slider] += dpos;
							mfd->lvG[slider] += dpos;
							mfd->lvB[slider] += dpos;
							SendMessage((HWND)lParam, TBM_SETPOS, (WPARAM)TRUE, -mfd->lvB[slider]);
						}
						break;
					}
				}
				else {
						vv = -SendMessage((HWND)lParam, TBM_GETPOS, 0, 0);
						switch (mfd->nmfunc) {
						case IDC_RC:
							if (mfd->wavesl) {
								switch ((int)LOWORD(wParam)) {
								case SB_THUMBTRACK: {
									dpos = vv - mfd->lvR[slider];
									int pos;
									for (i = slider + 1; i < slider + 4 && i < 17; i++)
									{
										pos = mfd->lvR[i] + (int)(dpos * cos(M_PI * (float)((i - slider) / 8.0)));
										norm1h(&pos);
										SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -pos);
									}
									for (i = slider - 1; i > slider - 4 && i >= 0; i--)
									{
										pos = mfd->lvR[i] + (int)(dpos * cos(M_PI * (float)((slider - i) / 8.0)));
										norm1h(&pos);
										SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -pos);
									}
													}
									break;
								case SB_THUMBPOSITION: {
									for (i = slider; i < slider + 4 && i < 17; i++)
									{
										mfd->lvR[i] = -SendMessage(vhWnd[i], TBM_GETPOS, 0, 0);
									}
									for (i = slider - 1; i > slider - 4 && i >= 0; i--)
									{
										mfd->lvR[i] = -SendMessage(vhWnd[i], TBM_GETPOS, 0, 0);
									}
													}
									break;
								default: {
									mfd->lvR[slider] = vv;
										 }
									break;
								}
							}
							else mfd->lvR[slider] = vv;
							break;
						case IDC_GC:
							if (mfd->wavesl) {
								switch ((int)LOWORD(wParam)) {
								case SB_THUMBTRACK: {
									dpos = vv - mfd->lvG[slider];
									int pos;
									for (i = slider + 1; i < slider + 4 && i < 17; i++)
									{
										pos = mfd->lvG[i] + (int)(dpos * cos(M_PI * (float)((i - slider) / 8.0)));
										norm1h(&pos);
										SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -pos);
									}
									for (i = slider - 1; i > slider - 4 && i >= 0; i--)
									{
										pos = mfd->lvG[i] + (int)(dpos * cos(M_PI * (float)((slider - i) / 8.0)));
										norm1h(&pos);
										SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -pos);
									}
													}
									break;
								case SB_THUMBPOSITION: {
									for (i = slider; i < slider + 4 && i < 17; i++)
									{
										mfd->lvG[i] = -SendMessage(vhWnd[i], TBM_GETPOS, 0, 0);
									}
									for (i = slider - 1; i > slider - 4 && i >= 0; i--)
									{
										mfd->lvG[i] = -SendMessage(vhWnd[i], TBM_GETPOS, 0, 0);
									}
													}
									break;
								default: {
									mfd->lvG[slider] = vv;
										 }
									break;
								}
							}
							else mfd->lvG[slider] = vv;
							break;
						case IDC_BC:
							if (mfd->wavesl) {
								switch ((int)LOWORD(wParam)) {
								case SB_THUMBTRACK: {
									dpos = vv - mfd->lvB[slider];
									int pos;
									for (i = slider + 1; i < slider + 4 && i < 17; i++)
									{
										pos = mfd->lvB[i] + (int)(dpos * cos(M_PI * (float)((i - slider) / 8.0)));
										norm1h(&pos);
										SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -pos);
									}
									for (i = slider - 1; i > slider - 4 && i >= 0; i--)
									{
										pos = mfd->lvB[i] + (int)(dpos * cos(M_PI * (float)((slider - i) / 8.0)));
										norm1h(&pos);
										SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -pos);
									}
													}
									break;
								case SB_THUMBPOSITION: {
									for (i = slider; i < slider + 4 && i < 17; i++)
									{
										mfd->lvB[i] = -SendMessage(vhWnd[i], TBM_GETPOS, 0, 0);
									}
									for (i = slider - 1; i > slider - 4 && i >= 0; i--)
									{
										mfd->lvB[i] = -SendMessage(vhWnd[i], TBM_GETPOS, 0, 0);
									}
													}
									break;
								default: {
									mfd->lvB[slider] = vv;
										 }
									break;
								}
							}
							else mfd->lvB[slider] = vv;
							break;
						case IDC_SAT:
							if (mfd->wavesl) {
								switch ((int)LOWORD(wParam)) {
								case SB_THUMBTRACK: {
									dpos = vv - mfd->lvS[slider];
									int pos;
									for (i = slider + 1; i < slider + 4 && i < 17; i++)
									{
										pos = mfd->lvS[i] + (int)(dpos * cos(M_PI * (float)((i - slider) / 8.0)));
										norm1h(&pos);
										SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -pos);
									}
									for (i = slider - 1; i > slider - 4 && i >= 0; i--)
									{
										pos = mfd->lvS[i] + (int)(dpos * cos(M_PI * (float)((slider - i) / 8.0)));
										norm1h(&pos);
										SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, -pos);
									}
													}
									break;
								case SB_THUMBPOSITION: {
									for (i = slider; i < slider + 4 && i < 17; i++)
									{
										mfd->lvS[i] = -SendMessage(vhWnd[i], TBM_GETPOS, 0, 0);
									}
									for (i = slider - 1; i > slider - 4 && i >= 0; i--)
									{
										mfd->lvS[i] = -SendMessage(vhWnd[i], TBM_GETPOS, 0, 0);
									}
													}
									break;
								default: {
									mfd->lvS[slider] = vv;
										 }
									break;
								}
							}
							else mfd->lvS[slider] = vv;
							break;
						}
				}
				setFlg(mfd);
				mfd->ifp->RedoFrame();
				break;
			}
	}

	return FALSE;
}

int ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	MyFilterData mfd_old = *mfd;
	int ret;

	mfd->ifp = fa->ifp;

	if (DialogBoxParam(fa->filter->module->hInstModule, MAKEINTRESOURCE(IDD_FILTER),
		hwnd, ConfigDlgProc, (LPARAM)mfd))
	{
		*mfd = mfd_old;
		ret = TRUE;
	}
    else
	{
		ret = FALSE;
	}
	return(ret);
}

void StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
}
 

void setFlg ( MyFilterData *mfd ) {
	int i;
	mfd->rgbFlg = false;
	for (i = 0; i < 17; i++) {
		mfd->rgbFlg = mfd->rgbFlg || mfd->lvR[i] || mfd->lvG[i] || mfd->lvB[i];
	}
	mfd->gamFlg = mfd->gamR || mfd->gamG || mfd->gamB;
	mfd->satFlg = false;
	for (i = 0; i < 17; i++) {
		mfd->satFlg = mfd->satFlg || mfd->lvS[i];
	}

		switch (mfd->nmfunc) {
		case IDC_RC:
		case IDC_GC:
		case IDC_BC:
		for (i = 0; i < 256; i++) {
			long ip = i >> 4;
			long upm1 = i - ip * 16;
			long upm2 = (ip + 1) * 16 - i;
			mfd->incr[i] = (mfd->lvR[ip] * upm2 + mfd->lvR[ip + 1] * upm1) >> 4;
			mfd->incg[i] = (mfd->lvG[ip] * upm2 + mfd->lvG[ip + 1] * upm1) >> 4;
			mfd->incb[i] = (mfd->lvB[ip] * upm2 + mfd->lvB[ip + 1] * upm1) >> 4;

			if (mfd->gamR <= 0) mfd->rtbl[i] = (unsigned char)(255 * pow(double(i+1)/256 , double(mfd->gamR * mfd->gamR)/1000 + 1));
			else mfd->rtbl[i] = (unsigned char)(255 * pow(double(i+1)/256 , double(100 - mfd->gamR)/100));
			if (mfd->gamG <= 0) mfd->gtbl[i] = (unsigned char)(255 * pow(double(i+1)/256 , double(mfd->gamG * mfd->gamG)/1000 + 1));
			else mfd->gtbl[i] = (unsigned char)(255 * pow(double(i+1)/256 , double(100 - mfd->gamG)/100));
			if (mfd->gamB <= 0) mfd->btbl[i] = (unsigned char)(255 * pow(double(i+1)/256 , double(mfd->gamB * mfd->gamB)/1000 + 1));
			else mfd->btbl[i] = (unsigned char)(255 * pow(double(i+1)/256 , double(100 - mfd->gamB)/100));
		}
			break;
		case IDC_SAT:
			for (i = 0; i < 256; i++) {
			long ip = i >> 4;
			long upm1 = i - ip*16;
			long upm2 = (ip + 1)*16 - i;
				mfd->incs[i] = (mfd->lvS[ip] * upm2 + mfd->lvS[ip + 1] * upm1) >> 4;
				mfd->incs[i] = 100 + mfd->incs[i] * 2 + abs(mfd->incs[i]);
				if (mfd->incs[i] < 0) mfd->incs[i] = 0;
			}
			mfd->satp = 100 + mfd->gamS * 2 + abs(mfd->gamS);
			if (mfd->satp < 0) mfd->satp = 0;
			break;
	}
}


int normpos( int a, int b, int c, int dpos) {
int ret;
if (dpos >0) {
		ret = a;
		if (b > ret) ret = b;
		if (c > ret) ret = c;
		if ((ret + dpos) < 101) ret = dpos;
		else ret = 100 - ret;
}
else {
		ret = a;
		if (b < ret) ret = b;
		if (c < ret) ret = c;
		if ((ret + dpos) > -101) ret = dpos;
		else ret = -100 - ret;
}
return(ret);
}

inline void normrgb(long *r, long *g, long *b) {
	if (*r < 0) *r = 0;
	if (*r > 255) *r = 255;
	if (*g < 0) *g = 0;
	if (*g > 255) *g = 255;
	if (*b < 0) *b = 0;
	if (*b > 255) *b = 255;
}

inline void gamma(MyFilterData *mfd, long *r, long *g, long *b) {
	*r = mfd->rtbl[*r];
	*g = mfd->gtbl[*g];
	*b = mfd->btbl[*b];
}

inline void rgbdo(MyFilterData *mfd, long *r, long *g, long *b) {
	long uold, u, up;
	if (mfd->separat) {
		*r += mfd->incr[*r];
		*g += mfd->incg[*g];
		*b += mfd->incb[*b];
		normrgb(r, g, b);
	}
	else {
		if (mfd->aveLum) {
		uold = *r + *g + *b;
		u = (uold)/3;
		}
		else {
			u = mmaxc(r,g,b);
			uold = u * 3;
		}
		*r += mfd->incr[u];
		*g += mfd->incg[u];
		*b += mfd->incb[u];
		normrgb(r, g, b);

		if (mfd->keepL) {
			if (mfd->aveLum) up = *r + *g + *b;
			else up = mmaxc(r,g,b) * 3;
			if (up > 0) {
				*r = (uold * *r)/up;
				*g = (uold * *g)/up;
				*b = (uold * *b)/up;
				normrgb(r, g, b);
			}
		}
	}
}

inline void satur(MyFilterData *mfd, long *r, long *g, long *b) {
	long u;
	if (mfd->aveLum || mfd->separat) u = (*r + *g + *b)/3;
	else u = mmaxc(r,g,b);
	*r = u + ((*r - u) * mfd->incs[u])/100;
	*g = u + ((*g - u) * mfd->incs[u])/100;
	*b = u + ((*b - u) * mfd->incs[u])/100;
	normrgb(r, g, b);
}

inline void satpr(MyFilterData *mfd, long *r, long *g, long *b) {
	long u;
	if (mfd->aveLum || mfd->separat) u = (*r + *g + *b)/3;
	else u = mmaxc(r,g,b);
	*r = u + ((*r - u) * mfd->satp)/100;
	*g = u + ((*g - u) * mfd->satp)/100;
	*b = u + ((*b - u) * mfd->satp)/100;
	normrgb(r, g, b);
}
UCHAR valtoch(int v, int ch) {
	UCHAR ret = 0;
	switch (ch) {
	case 1:
		if (v < 0) ret = 110;
		else ret = 112;
		return ret;
	case 2:
		ret = (UCHAR)(abs(v) + 128);
		return ret;
	}
	return ret;
}

int chtoval(UCHAR ch1, UCHAR ch2) {
	int ret = 0;
	switch (ch1) {
	case 110:
		ret = -(int)(ch2 - 128);
		return ret;
	case 112:
		ret = (int)(ch2 - 128);
		return ret;
	}
	return ret;
}
inline long mmaxc(long *r, long *g, long *b) {
long ret = *r;
if (*g > ret) ret = *g;
if (*b > ret) ret = *b;
return ret;
}

inline void norm1h(int *v) {
	if (*v > 100) *v = 100;
	if (*v < -100) *v = -100;
}